// Headers
#include <tchar.h>
#define _WIN32_WINNT 0x0500
#include <windows.h>

// FlashPlayerControl
#include "../../../FlashPlayerControl/include/FlashPlayerControl.h"
#pragma comment(lib, "../../../FlashPlayerControl/lib/FlashPlayerControl.lib")

// Direct 3D
#include <d3d8.h>
#include <d3dx8math.h>
#pragma comment(lib, "d3d8.lib")
#pragma comment(lib, "d3dx8.lib")

// Defines
#define DEF_MAIN_WND_CLASS_NAME				(_T("{E6AA4E57-62BE-4bf3-AF8C-7EE68FA9D2A4}"))

// Forward declarations
LRESULT CALLBACK WndProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
void SetupMatrices();
void RenderScene();
LPDIRECT3DTEXTURE8 CreateTextureFromFlash();

#define D3DFVF_MYVERTEX (D3DFVF_XYZ | D3DFVF_DIFFUSE | D3DFVF_TEX1)

struct MYVERTEX
{
	FLOAT x, y ,z;
	D3DCOLOR color;
	FLOAT tu, tv;
};

#define DEF_WIDTH			(128)
#define DEF_HEIGHT			(128)

#define DEF_MAIN_WND_WIDTH  (480)
#define DEF_MAIN_WND_HEIGHT (480)

// Globals
LPDIRECT3D8 g_PDIRECT3D8 = NULL;
LPDIRECT3DDEVICE8 g_PDIRECT3DDEVICE8 = NULL;
LPDIRECT3DVERTEXBUFFER8 g_PDIRECT3DVERTEXBUFFER8 = NULL;
D3DTEXTUREFILTERTYPE CurrentFilter;
HWND g_hwndFlashPlayerControl = NULL;

// Entry point
int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE, LPSTR, int nCmdShow)
{
	HWND hWnd;

	WNDCLASS wndclass = { 0 };
	wndclass.lpfnWndProc = &WndProc;
	wndclass.hInstance = hInstance;
	wndclass.hIcon = LoadIcon(NULL, IDI_APPLICATION);
	wndclass.hCursor = LoadCursor(NULL, IDC_ARROW);
	wndclass.hbrBackground = GetSysColorBrush(COLOR_APPWORKSPACE);
	wndclass.lpszClassName = DEF_MAIN_WND_CLASS_NAME;

	if (!RegisterClass(&wndclass))
		return 0;

	hWnd = 
		CreateWindow(DEF_MAIN_WND_CLASS_NAME, 
					 _T("FlashPlayerControl Example - FlashTexture"), 
					 WS_SYSMENU | WS_OVERLAPPED | WS_MINIMIZEBOX, 
					 GetSystemMetrics(SM_CXSCREEN) / 2 - DEF_MAIN_WND_WIDTH / 2, 
					 GetSystemMetrics(SM_CYSCREEN) / 2 - DEF_MAIN_WND_HEIGHT / 2, 
					 DEF_MAIN_WND_WIDTH, 
					 DEF_MAIN_WND_HEIGHT, 
					 NULL, 
					 NULL, 
					 hInstance, 
					 NULL);

	// Register FlashPlayerControl window class name
    if (!RegisterFlashWindowClass())
    {
        MessageBox(NULL, _T("RegisterFlashWindowClass() failed"), _T("Error"), MB_OK | MB_ICONSTOP);
        return FALSE;
    }

	// Create FlashPlayerControl as a hidden child control 
	// in transparent mode (see style FPCS_TRANSPARENT)
	g_hwndFlashPlayerControl = 
		CreateWindow(WC_FLASH, 
					 NULL, 
					 WS_CHILD | FPCS_TRANSPARENT, 
					 0, 
					 0, 
					 DEF_WIDTH, 
					 DEF_HEIGHT, 
					 hWnd, 
					 NULL, 
					 NULL, 
					 NULL);

	// Load movie from resource
	FPCPutMovieFromResource(g_hwndFlashPlayerControl, NULL, _T("SWF1"), _T("SWF"));
	// Play
	FPC_Play(g_hwndFlashPlayerControl);

	CurrentFilter = D3DTEXF_NONE;

	// Direct 3D Initialization
	{
		D3DPRESENT_PARAMETERS present_parameters = { 0 };

		if (NULL == (g_PDIRECT3D8 = Direct3DCreate8(D3D_SDK_VERSION)))
			return FALSE;

		D3DDISPLAYMODE display_mode;

		if (FAILED(g_PDIRECT3D8->GetAdapterDisplayMode(D3DADAPTER_DEFAULT, &display_mode)))
			return FALSE;

		present_parameters.Windowed = TRUE;
		present_parameters.BackBufferFormat = display_mode.Format;
		present_parameters.SwapEffect = D3DSWAPEFFECT_DISCARD;
		present_parameters.AutoDepthStencilFormat = D3DFMT_D16;
		present_parameters.EnableAutoDepthStencil = TRUE;

		if (FAILED(g_PDIRECT3D8->CreateDevice(D3DADAPTER_DEFAULT, 
											  D3DDEVTYPE_HAL, 
											  hWnd, 
											  D3DCREATE_SOFTWARE_VERTEXPROCESSING, 
											  &present_parameters, 
											  &g_PDIRECT3DDEVICE8)))
			if (FAILED(g_PDIRECT3D8->CreateDevice(D3DADAPTER_DEFAULT, 
												  D3DDEVTYPE_REF, 
												  hWnd, 
												  D3DCREATE_SOFTWARE_VERTEXPROCESSING, 
												  &present_parameters, 
												  &g_PDIRECT3DDEVICE8)))
				return FALSE;

		g_PDIRECT3DDEVICE8->SetRenderState(D3DRS_ZENABLE, TRUE);
		g_PDIRECT3DDEVICE8->SetRenderState(D3DRS_CULLMODE, D3DCULL_CW);
		g_PDIRECT3DDEVICE8->SetRenderState(D3DRS_LIGHTING, FALSE);
	}

	// Scene
	{
		const float a = 5;

		const MYVERTEX A = { -a / 2,  a / 2, 0, 0x00ffffff, 0, 1 };
		const MYVERTEX B = { -a / 2, -a / 2, 0, 0x00ffffff, 0, 0 };
		const MYVERTEX C = {  a / 2, -a / 2, 0, 0x00ffffff, 1, 0 };
		const MYVERTEX D = {  a / 2,  a / 2, 0, 0x00ffffff, 1, 1 };
		const MYVERTEX E = { 0, 0, a / sqrt(2.0f), 0x00ffffff, .5, .5};

		const MYVERTEX Vertices[] = 
		{
			E,	A,	D,
			E,	B,	A,
			E,	C,	B,
			E,	D,	C,
		};

		if (FAILED(g_PDIRECT3DDEVICE8->CreateVertexBuffer(sizeof(Vertices), 
														  0, 
														  D3DFVF_MYVERTEX, 
														  D3DPOOL_DEFAULT, 
														  &g_PDIRECT3DVERTEXBUFFER8)))
			return FALSE;

		VOID* pVertices;

		if (FAILED(g_PDIRECT3DVERTEXBUFFER8->Lock(0, sizeof(Vertices), (BYTE**)&pVertices, 0)))
			return FALSE;

		CopyMemory(pVertices, Vertices, sizeof(Vertices));

		g_PDIRECT3DVERTEXBUFFER8->Unlock();
	}

	ShowWindow(hWnd, nCmdShow);
	UpdateWindow(hWnd);

	MSG msg = { 0 };

	while (WM_QUIT != msg.message)
		if(PeekMessage(&msg, NULL, 0, 0, PM_REMOVE))
		{
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}
		else
			RenderScene();

	// Cleanup
	if (NULL != g_PDIRECT3DVERTEXBUFFER8)
		g_PDIRECT3DVERTEXBUFFER8->Release();

	if (NULL != g_PDIRECT3DDEVICE8)
		g_PDIRECT3DDEVICE8->Release();

	if (NULL != g_PDIRECT3D8)
		g_PDIRECT3D8->Release();

	//
	UnregisterFlashWindowClass();

	UnregisterClass(DEF_MAIN_WND_CLASS_NAME, hInstance);

	return 0;
}

LRESULT CALLBACK WndProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	switch (uMsg)
	{
		case WM_DESTROY:
		{
			PostQuitMessage(0);
			return 0;
		}
	}

	return DefWindowProc(hWnd, uMsg, wParam, lParam);
}

void SetupMatrices()
{
	static DWORD g_dwStartTime = 0;

	if (0 == g_dwStartTime)
		g_dwStartTime = GetTickCount();

	D3DXMATRIX WorldMatrix;
	D3DXMatrixIdentity(&WorldMatrix);
	D3DXMatrixRotationZ(&WorldMatrix, (GetTickCount() - g_dwStartTime) / 1024.0f);
	g_PDIRECT3DDEVICE8->SetTransform(D3DTS_WORLD, &WorldMatrix);

	D3DXMATRIX ViewMatrix;
	D3DXMatrixLookAtLH(&ViewMatrix, 
					   &D3DXVECTOR3(5.0f, 5.0f, 6.5f), 
					   &D3DXVECTOR3(0.0f, 0.0f, 1.0f), 
					   &D3DXVECTOR3(0.0f, 0.0f, 1.0f));
	g_PDIRECT3DDEVICE8->SetTransform(D3DTS_VIEW, &ViewMatrix);

	D3DXMATRIX ProjectionMatrix;
	D3DXMatrixPerspectiveFovLH(&ProjectionMatrix, D3DX_PI / 3, 1.0f, 1.0f, 100.0f);
	g_PDIRECT3DDEVICE8->SetTransform(D3DTS_PROJECTION, &ProjectionMatrix );
}

void RenderScene()
{
	g_PDIRECT3DDEVICE8->Clear(0, 
							  NULL, 
							  D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER, 
							  D3DCOLOR_RGBA(100, 100, 100, 50), 
							  1.0, 
							  0);

	g_PDIRECT3DDEVICE8->BeginScene();

	SetupMatrices();

	LPDIRECT3DTEXTURE8 pTexture = CreateTextureFromFlash();

	g_PDIRECT3DDEVICE8->SetTexture(0, pTexture);
	g_PDIRECT3DDEVICE8->SetTextureStageState(0, D3DTSS_MAGFILTER, CurrentFilter);
	g_PDIRECT3DDEVICE8->SetTextureStageState(0, D3DTSS_MINFILTER, CurrentFilter);
	g_PDIRECT3DDEVICE8->SetTextureStageState(0, D3DTSS_MIPFILTER, CurrentFilter);

	g_PDIRECT3DDEVICE8->SetStreamSource(0, g_PDIRECT3DVERTEXBUFFER8, sizeof(MYVERTEX));
	g_PDIRECT3DDEVICE8->SetVertexShader(D3DFVF_MYVERTEX);
	//   
	g_PDIRECT3DDEVICE8->DrawPrimitive(D3DPT_TRIANGLELIST, 0, 4);

	g_PDIRECT3DDEVICE8->EndScene();

	g_PDIRECT3DDEVICE8->Present(NULL, NULL, NULL, NULL);

	if (NULL != pTexture)
		pTexture->Release();
}

LPDIRECT3DTEXTURE8 CreateTextureFromFlash()
{
	LPDIRECT3DTEXTURE8 pTexture = NULL;

	D3DXCreateTexture(g_PDIRECT3DDEVICE8, 
					  DEF_WIDTH, 
					  DEF_HEIGHT, 
					  1, 
					  0, 
					  D3DFMT_A8R8G8B8, 
					  D3DPOOL_MANAGED, 
					  &pTexture);

	if (NULL == pTexture)
		return pTexture;

	D3DSURFACE_DESC d3dsd;
	pTexture->GetLevelDesc(0, &d3dsd);

	// If size is changed...
	MoveWindow(g_hwndFlashPlayerControl, 0, 0, d3dsd.Width, d3dsd.Height, TRUE);

	// Size
	SIZE size = { d3dsd.Width, d3dsd.Height };

	D3DLOCKED_RECT rcLockedRect = { 0 };
	RECT rc = { 0, 0, size.cx, size.cy };

	pTexture->LockRect(0, &rcLockedRect, &rc, D3DLOCK_DISCARD);
	{
		SFPCGetFrameBitmap FPCGetFrameBitmap = { 0 };
		::SendMessage(g_hwndFlashPlayerControl, FPCM_GET_FRAME_BITMAP, 0, (LPARAM)&FPCGetFrameBitmap);
		HBITMAP hBitmap = FPCGetFrameBitmap.hBitmap;
		{
			BITMAP bmp_info;
			GetObject(hBitmap, sizeof(bmp_info), &bmp_info);

			BYTE* pTextureBits = (BYTE*)rcLockedRect.pBits;
			DWORD* pBitmapBits = (DWORD*)bmp_info.bmBits;

			BYTE* pLineTextureBits = pTextureBits;

			for (int j = 0; j < size.cy; j++)
			{
				DWORD* pPixels = (DWORD*)pLineTextureBits;

				for (int i = 0; i < size.cx; i++)
				{
					*pPixels = *pBitmapBits;

					pPixels++;
					pBitmapBits++;
				}

				pLineTextureBits += rcLockedRect.Pitch;
			}
		}
		DeleteObject(hBitmap);
	}
	pTexture->UnlockRect(0);

	return pTexture;
}
